/*
   Kickshaw - A Menu Editor for Openbox

   Copyright (c) 2010–2021        Marcus Schätzle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along 
   with Kickshaw. If not, see http://www.gnu.org/licenses/.
*/

#ifndef __definitions_and_enumerations_h
#define __definitions_and_enumerations_h

#include <stdbool.h>

// possible edits done to the menu during loading process
enum { ICON_INFO_CORRECTED, INVALID_OPTION_VALUE_CORRECTED, INVISIBLE_ELEMENT_VISUALIZED_OR_REMOVED, DEPRECATED_EXE_CMD_CVTD, 
       NUMBER_OF_LOADING_PROCESS_EDIT_TYPES };
// action option combo
enum { ACTION_OPTION_COMBO_ITEM, NUMBER_OF_ACTION_OPTION_COMBO_ELEMENTS };
// new action option grid/table widgets
enum { INSIDE_MENU_LABEL, INSIDE_MENU_CHECK_BUTTON, INCLUDING_ACTION_LABEL, INCLUDING_ACTION_CHECK_BUTTON, 
       NEW_ACTION_OPTION_COMBO_BOX, ACTION_OPTION_DONE, ACTION_OPTION_CANCEL, NUMBER_OF_NEW_ACTION_OPTION_WIDGETS };
// actions
enum { EXECUTE, EXIT, RECONFIGURE, RESTART, SESSIONLOGOUT, NUMBER_OF_ACTIONS };
// add buttons
enum { MENU, PIPE_MENU, ITEM, SEPARATOR, ACTION_OR_OPTION, NUMBER_OF_ADD_BUTTONS };
// ancestor visibility
enum { NONE_OR_VISIBLE_ANCESTOR, INVISIBLE_ANCESTOR, INVISIBLE_ORPHANED_ANCESTOR };
// columns
enum { COL_MENU_ELEMENT, COL_TYPE, COL_VALUE, COL_MENU_ID, COL_EXECUTE, COL_ELEMENT_VISIBILITY, NUMBER_OF_COLUMNS };
// entry fields
enum { MENU_ELEMENT_OR_VALUE_ENTRY, ICON_PATH_ENTRY, MENU_ID_ENTRY, EXECUTE_ENTRY, NUMBER_OF_ENTRY_FIELDS };
// execute options
enum { PROMPT, COMMAND, STARTUPNOTIFY, SN_OR_PROMPT = 2, NUMBER_OF_EXECUTE_OPTS };
// find buttons
enum { CLOSE, BACK, FORWARD, NUMBER_OF_FIND_ENTRY_BUTTONS };
// special options for find
enum { MATCH_CASE, REGULAR_EXPRESSION, WHOLE_WORD, NUMBER_OF_SPECIAL_OPTIONS };
// find and replace entry fields
enum { REPLACE_ENTRY, WITH_ENTRY, NUMBER_OF_FR_ENTRY_FIELDS };
// invalid icon images
enum { INVALID_PATH_ICON, INVALID_FILE_ICON, NUMBER_OF_INVALID_ICON_IMGS };
// invalid icon image statuses
enum { NONE_OR_NORMAL, INVALID_PATH, INVALID_FILE };
// popup menus
enum { FILE_MENU, EDIT_MENU, SEARCH_MENU, VIEW_MENU, OPTIONS_MENU, HELP_MENU, NUMBER_OF_MENUS };
// popup menu items file
enum { M_NEW, M_OPEN, M_SAVE, M_SAVE_AS, M_QUIT, NUMBER_OF_FILE_MENU_ITEMS };
// popup menu items edit
enum { M_UNDO, M_REDO, M_MOVE_TOP, M_MOVE_UP, M_MOVE_DOWN, M_MOVE_BOTTOM, 
       M_REMOVE, M_REMOVE_ALL_CHILDREN, M_VISUALIZE, M_VISUALIZE_RECURSIVELY, NUMBER_OF_EDIT_MENU_ITEMS };
// popup menu items find (& replace)
enum { M_FIND, M_FIND_AND_REPLACE, NUMBER_OF_SEARCH_MENU_ITEMS };
// popup menu items view
enum { M_EXPAND_ALL_NODES, M_COLLAPSE_ALL_NODES, M_SHOW_MENU_ID_COL, M_SHOW_EXECUTE_COL, M_SHOW_ELEMENT_VIS_COL, 
       M_SHOW_ICONS, M_HIGHLIGHT_SEPARATORS, M_SHOW_TREE_LINES, M_SHOW_GRID, NUMBER_OF_VIEW_MENU_ITEMS };
// popup menu element visibity submenu items
enum { M_SHOW_ELEMENT_VISIBILITY_COL_ACTVTD, M_SHOW_ELEMENT_VISIBILITY_COL_KEEP_HIGHL, NUMBER_OF_VIEW_MENU_VISIBILITY_ITEMS };
// popup menu grid submenu items
enum { M_NO_GRID_LINES, M_SHOW_GRID_HOR, M_SHOW_GRID_VER, M_BOTH, NUMBER_OF_VIEW_MENU_GRID_ITEMS };
// popup menu items options
enum { M_CREATE_BACKUP_BEFORE_OVERWRITING_MENU, M_USE_TABS_FOR_INDENTATIONS, M_KEEP_ROOT_MENU_SEPARATE, 
       M_SORT_EXECUTE_AND_STARTUPN_OPTIONS, M_NOTIFY_ABOUT_EXECUTE_OPT_CONVERSIONS, M_USE_HEADER_BAR, M_SHOW_MENU_BUTTON, 
       NUMBER_OF_OPTIONS_MENU_ITEMS };
// popup menu items help
enum { M_HINTS, M_SHORTCUTS, M_REPORT_ISSUE, M_ABOUT, NUMBER_OF_HELP_MENU_ITEMS };
// move row
enum { MOVE_TOP, MOVE_UP, MOVE_DOWN, MOVE_BOTTOM, NUMBER_OF_MOVEMENT_BUTTONS };
// startupnotify options
enum { ENABLED, NAME, WM_CLASS, ICON, NUMBER_OF_STARTUPNOTIFY_OPTS };
// toolbar buttons
enum { TB_BUTTON_NEW, TB_BUTTON_OPEN, TB_BUTTON_SAVE, TB_BUTTON_SAVE_AS, TB_BUTTON_UNDO, TB_BUTTON_REDO, 
       TB_BUTTON_FIND, TB_BUTTON_EXPAND_ALL_NODES, TB_BUTTON_COLLAPSE_ALL_NODES, NUMBER_OF_TB_BUTTONS };
// treestore elements
enum { TS_ICON_IMG, TS_ICON_IMG_STATUS, TS_ICON_MODIFICATION_TIME, TS_ICON_PATH, TS_MENU_ELEMENT, 
       TS_TYPE, TS_VALUE, TS_MENU_ID, TS_EXECUTE, TS_ELEMENT_VISIBILITY, NUMBER_OF_TS_ELEMENTS };
// text fields
enum { ICON_PATH_TXT, MENU_ELEMENT_TXT, TYPE_TXT, VALUE_TXT, MENU_ID_TXT, 
       EXECUTE_TXT, ELEMENT_VISIBILITY_TXT, NUMBER_OF_TXT_FIELDS };
// renderer for treeview
enum { TXT_RENDERER, EXCL_TXT_RENDERER, PIXBUF_RENDERER, BOOL_RENDERER, NUMBER_OF_RENDERERS };
// hint texts
enum { MENU_FILE_HINTS, WORKING_WITH_KS_HINTS, MISCELLANEOUS_HINTS, NUMBER_OF_HINT_TXTS };

G_DEFINE_AUTOPTR_CLEANUP_FUNC(PangoFontDescription, pango_font_description_free)
G_DEFINE_AUTOPTR_CLEANUP_FUNC(PangoLayout, g_object_unref)
G_DEFINE_AUTOPTR_CLEANUP_FUNC(PangoTabArray, pango_tab_array_free)

#define KICKSHAW_VERSION "1.0.54"
#define TREEVIEW_COLUMN_OFFSET NUMBER_OF_TS_ELEMENTS - NUMBER_OF_COLUMNS
#define FREE_AND_REASSIGN(string, new_value) { g_free (string); string = new_value; }
#define NOT_NULL_AND_NOT_EMPTY(string) (string && *string)
#define STREQ(string1, string2) (g_strcmp0 ((string1), (string2)) == 0)
#define FOREACH_IN_ARRAY(array, counter) for (counter = 0; counter < sizeof (array) / sizeof *(array); counter++)
#define FOREACH_IN_LIST(list, counter) for (counter = list; counter != NULL; counter = counter->next)

#if !(GTK_CHECK_VERSION(3,24,0))
G_DEFINE_AUTOPTR_CLEANUP_FUNC(GResource, g_resource_unref)
G_DEFINE_AUTOPTR_CLEANUP_FUNC(GtkTreePath, gtk_tree_path_free)
#endif

/* bool should be used instead of gboolean here.
   gboolean is a typedef for int, so when TRUE is assigned to a gboolean bit field 
   this can result in a compiler warning like the following: 
   overflow in conversion from ‘int’ to ‘signed char:1’ changes value from ‘1’ to ‘-1’ */
typedef struct {
    bool at_least_one_is_expanded        : 1;
    bool at_least_one_imd_ch_is_expanded : 1;
    bool at_least_one_is_collapsed       : 1;
} expansion_statuses;

typedef struct {
    GtkApplication *app;

    GtkWidget *window;
    GtkWidget *header_bar;
    GtkWidget *sub_box;

    GtkWidget *basename_label;
    GtkWidget *subtitle_label;

    GtkWidget *hints_window;

    GMenu *popup_menu;
    GMenuItem *MBut_menu_items[NUMBER_OF_MENUS];

    GSimpleAction *file_actions[NUMBER_OF_FILE_MENU_ITEMS];
    GSimpleAction *edit_actions[NUMBER_OF_EDIT_MENU_ITEMS];
    GSimpleAction *view_actions[NUMBER_OF_VIEW_MENU_ITEMS];
    GSimpleAction *view_action_visibility, *view_action_highlighting;
    GSimpleAction *options_actions[NUMBER_OF_OPTIONS_MENU_ITEMS];

    const gchar *grid_submenu_item_txts[NUMBER_OF_VIEW_MENU_GRID_ITEMS];
    const gchar *grid_states[NUMBER_OF_VIEW_MENU_GRID_ITEMS];

    const gchar *edit_simple_action_txts[NUMBER_OF_EDIT_MENU_ITEMS];

    GMenuItem *MBut_options_menu_items[NUMBER_OF_OPTIONS_MENU_ITEMS];

    GtkWidget *MBar_menus[NUMBER_OF_MENUS];
    GtkWidget *MBar_menu_items[NUMBER_OF_MENUS];
    GtkWidget *MBar_file_menu_items[NUMBER_OF_EDIT_MENU_ITEMS];
    GtkWidget *MBar_edit_menu_items[NUMBER_OF_EDIT_MENU_ITEMS];
    GtkWidget *MBar_view_menu_items[NUMBER_OF_VIEW_MENU_ITEMS];
    GtkWidget *MBar_options_menu_items[NUMBER_OF_OPTIONS_MENU_ITEMS];
    GtkWidget *MBar_view_menu_visibility_items[NUMBER_OF_VIEW_MENU_VISIBILITY_ITEMS];
    GtkWidget *MBar_view_menu_grid_items[NUMBER_OF_VIEW_MENU_GRID_ITEMS];

    GtkWidget *tb_buttons[NUMBER_OF_TB_BUTTONS];
    GtkWidget *tb_button_quit;

    GtkWidget *button_grid;
    GtkWidget *add_image;
    GtkWidget *bt_bar_label;
    GtkWidget *bt_add[NUMBER_OF_ADD_BUTTONS];
    GtkWidget *bt_add_action_option_label;
    GtkWidget *bt_remove;
    GtkWidget *bt_movements[NUMBER_OF_MOVEMENT_BUTTONS];

    GtkWidget *enter_values_box, *enter_values_label;

    GtkWidget *action_option_grid, *new_action_option_grid_or_table;
    GtkWidget *new_action_option_widgets[NUMBER_OF_NEW_ACTION_OPTION_WIDGETS];
    GtkListStore *action_option_combo_box_liststore;
    GtkTreeModel *action_option_combo_box_model;
    GtkWidget *mandatory;
    GString *mandatory_label_txt;
    GtkWidget *options_grid, *suboptions_grid;
    const gchar *options_label_txts[NUMBER_OF_EXECUTE_OPTS];
    GtkWidget *options_labels[NUMBER_OF_EXECUTE_OPTS], *options_fields[NUMBER_OF_EXECUTE_OPTS];
    GtkCssProvider *execute_options_css_providers[NUMBER_OF_EXECUTE_OPTS - 1]; // Only "command" and "prompt".
    GtkWidget *suboptions_labels[NUMBER_OF_STARTUPNOTIFY_OPTS], *suboptions_fields[NUMBER_OF_STARTUPNOTIFY_OPTS];
    GtkCssProvider *suboptions_fields_css_providers[NUMBER_OF_STARTUPNOTIFY_OPTS - 1]; // Only "name", "wmclass" and "icon".

    const gchar *actions[NUMBER_OF_ACTIONS];
    const gchar *execute_options[NUMBER_OF_EXECUTE_OPTS];
    const gchar *execute_displayed_txts[NUMBER_OF_EXECUTE_OPTS];

    const gchar *startupnotify_options[NUMBER_OF_STARTUPNOTIFY_OPTS];
    const gchar *startupnotify_displayed_txts[NUMBER_OF_STARTUPNOTIFY_OPTS];

    GtkWidget *separator, *enter_values_buttons_grid;

    GtkWidget *find_grid;
    GtkWidget *find_entry_buttons[NUMBER_OF_FIND_ENTRY_BUTTONS];
    GtkWidget *find_entry;
    GtkCssProvider *find_entry_css_provider;
    GtkWidget *find_in_columns[NUMBER_OF_COLUMNS - 1], *find_in_all_columns;
    GtkCssProvider *find_in_columns_css_providers[NUMBER_OF_COLUMNS - 1], *find_in_all_columns_css_provider;
    GtkWidget *special_options[NUMBER_OF_SPECIAL_OPTIONS];
    GString *search_term;
    GList *rows_with_found_occurrences;

    GtkWidget *fr_dialog;
    GtkWidget *fr_entry_fields[NUMBER_OF_FR_ENTRY_FIELDS];
    GtkWidget *fr_in_columns[NUMBER_OF_COLUMNS - 1];
    GtkWidget *fr_in_all_columns;
    GtkWidget *fr_special_options[NUMBER_OF_SPECIAL_OPTIONS];
    GtkWidget *fr_replacements_done_label;
    GtkCssProvider *fr_in_columns_css_providers[NUMBER_OF_COLUMNS - 1], *fr_in_all_columns_css_provider;

    GtkWidget *text_views[NUMBER_OF_HINT_TXTS];

    GtkWidget *treeview;
    GtkTreeStore *treestore;
    GtkTreeModel *ts_model;
    GtkTreeIter iter;
    const gchar *column_header_txts[NUMBER_OF_COLUMNS];
    GtkTreeViewColumn *columns[NUMBER_OF_COLUMNS];
    GtkCellRenderer *renderers[NUMBER_OF_RENDERERS];

    gchar *txt_fields[NUMBER_OF_TXT_FIELDS];

    GtkTargetEntry enable_list[3];

    GtkWidget *entry_grid;
    GtkWidget *entry_labels[NUMBER_OF_ENTRY_FIELDS], *entry_fields[NUMBER_OF_ENTRY_FIELDS];
    GString *label_field_txt, *menu_ID_field_txt, *execute_field_txt, *prompt_field_txt, *command_field_txt, *mandatory_empty_string_txt;
    GtkWidget *double_menu_id_label;
    GString *double_menu_id_label_txt;
    GtkCssProvider *menu_element_or_value_entry_css_provider, *icon_path_entry_css_provider, 
                   *menu_id_css_provider, *execute_entry_css_provider;
    GtkWidget *icon_chooser, *remove_icon;

    GtkListStore *liststore;
    GtkTreeModel *ls_model;

    gchar *settings_file_actions;
    GPtrArray *settings_menu_items;

    GSList *source_paths;

    GSList *enter_values_user_settings;

    GtkWidget *statusbar;
    gboolean statusbar_msg_shown;

    GtkCssProvider *cm_css_provider;

    gboolean dark_theme;
    gboolean bg_image_or_gradient;
    gchar *red_hue, *blue_hue;

    gchar *icon_theme_name;
    gchar *font_desc;
    guint font_size;

    GdkPixbuf *invalid_icon_imgs[NUMBER_OF_INVALID_ICON_IMGS];

    gchar *filename;

    gchar *tmp_path;
    GFile *symlink_tmp_folder;

    gboolean restored_autosave_that_hasnt_been_saved_yet;

    GSList *menu_ids;
    GSList *rows_with_icons;

    gint number_of_undo_stack_items;
    gint pos_inside_undo_stack;
    gint last_save_pos;

    gboolean use_header_bar;
    gboolean show_menu_button;
    gboolean autosort_options;
    gboolean tabs_for_indentation;
    gboolean separate_root_menu;

    gboolean loading_process_edit_types[NUMBER_OF_LOADING_PROCESS_EDIT_TYPES];
    gboolean change_done;

    guint timeout_id;

    gulong handler_id_row_selected, 
           handler_id_action_option_combo_box, handler_id_action_option_button_clicked, 
           handler_id_show_or_hide_startupnotify_options, 
           handler_id_use_header_bar, 
           handler_id_show_menu_button, 
           handler_id_find_in_columns[NUMBER_OF_COLUMNS - 1], 
           handler_id_fr_in_columns[NUMBER_OF_COLUMNS - 1], 
           handler_id_entry_fields[NUMBER_OF_ENTRY_FIELDS], 
           handler_id_entry_field_changed, 
           handler_id_double_menu_id, 
           handler_id_including_action_check_button;
} ks_data;

#endif
